/*************************************************************************
/ RED.C
/ Implementa el nivel de red: interfaz entre el nivel de enlace y el
/ de transporte.
/*************************************************************************/

#include "red.h"


t_maquina *red[TAM_HASH];	          // tabla de maquinas conectadas

/*************************************************************************
FUNCIONALIDAD: Funcion hash que nos devuelve la XOR de todos los caracteres
			   que posea la cadena que estamos tratando, indicando una
			   entrada en la tablas de mquinas que estn disponibles en
			   la red.
PARAMETROS:
    cadena: cadena a la que se calcular el valor hash
DEVUELVE: Entrada en la tabla de mquinas.
/*************************************************************************/
int f_hash(char cadena[]) {

  int i,           // variable auxiliar
	  valor = 0;   // valor hash

  for (i = 0; i < strlen(cadena); i++)
	 valor ^= (int) cadena[i];

  valor %= TAM_HASH;
  return valor;
}


/*************************************************************************
FUNCIONALIDAD: Inserta una nueva maquina en la tabla de maquinas conectadas     segun
			   una funcion hash. NO COMPRUEBA QUE YA EXISTA, SIMPLEMENTE
			   AADE.
PARAMETROS:
    logico: nombre logico de la maquina
    fisico: direccin fisica de la maquina
DEVUELVE: Si se ha insertado, devuelve un 0. En caso contrario, se asume
		  la existencia de un error y devuelve -1.
/*************************************************************************/
int anadir_maquina(char logico[], char fisico[]){

  t_maquina *new_maq;
  int hash_valor;

  if ((new_maq = (t_maquina *) malloc (sizeof(t_maquina))) == NULL)
	 return (-1);

  memcpy(new_maq->logico, logico, TAM_MAX_NOMBRE_LOG);
  memcpy(new_maq->fisico, fisico, TAM_DIR);

  hash_valor = f_hash(logico);
  new_maq->siguiente = red[hash_valor];
  red[hash_valor] = new_maq;
  return (0);
}


/*************************************************************************
FUNCIONALIDAD: Comprueba si tenemos una maquina con el nombre nombre_logico
			   y nos devuelve su direccin fsica en caso que que exista.
PARAMETROS:
    nombre_logico: nombre logico de la mquina
    dir_fis: devuelve la direccin fsica de la mquina en caso de que
			  se encuentre una mquina con el nombre_logico
DEVUELVE: Si no la encuentra devuelve 0 (indeterminado el valor de dir_fis).
		  En caso de encontrarla, devuelve 1 y la direccin fsica en el
		  parametro correspondiente.
/*************************************************************************/
int buscar_maquina(char *nombre_logico, char *dir_fis){

   char valor_hash;      // Valor hash del nombre lgico
   t_maquina *aux_maq;   // variable auxiliar para buscar la mquina

   valor_hash = f_hash(nombre_logico);
   aux_maq = red[valor_hash];

   while ((aux_maq != NULL) && strcmp(aux_maq->logico, nombre_logico))
	  aux_maq = aux_maq->siguiente;
   if (aux_maq == NULL)
	   return (0);
   else
	   memcpy (dir_fis, aux_maq->fisico, TAM_DIR);
   return (1);
}


/*************************************************************************
FUNCIONALIDAD: Comprueba si tenemos una maquina con el nombre nombre_logico
			   y la direccion fisica dir_fis en la tabla de maquinas.
			   Si no existe se inserta.
PARAMETROS:
    nombre_logico: nombre lgico de la mquina
    dir_fis: direccin fisica de la mquina
    insertar: indica si hay que insertar la mquina definida por los
			   parmetros anteriores en el caso de que no haya ninguna
			   otra con su mismo nombre lgico y direccin fsica.
DEVUELVE:
   -1, si existi un error y no se ha podido efectuar la operacin
	0, si ya existe un usuario con el mismo nombre logico y diferente fisico
	1, si ya existe un usuario con el mismo nombre logico y fisico
	2, si no existe una entrada con el mismo nombre logico. Este es el
	   unico caso en que se inserta al nuevo usuario si el parametro
	   'insertar' esta a 1.
***********************************************************************/
int comprueba_maquina (char *nombre_logico, char *dir_fis, int insertar) {

   char valor_hash;       // Valor hash del parametro nombre logico
   t_maquina *aux_maq;    // variable auxiliar para buscar la mquina

   valor_hash = f_hash(nombre_logico);
   aux_maq = red[valor_hash];

   while ((aux_maq != NULL) && strcmp(aux_maq->logico, nombre_logico))
	  aux_maq = aux_maq->siguiente;
   if (aux_maq == NULL) {
	  if (insertar)
		 if (anadir_maquina(nombre_logico,dir_fis) == -1)
			return (-1);
	  return (2);
   }
   if (memcmp(dir_fis,aux_maq->fisico,TAM_DIR))
	  return (0);
   return(1);
}



/*************************************************************************
FUNCIONALIDAD: Busca una maquina en la red y si la encuentra, la elimina.
PARAMETROS:
    nombre_logico: nombre lgico de la mquina a eliminar
DEVUELVE:
	0, si no exista la mquina en la red
	1, si se elimin de la red
***********************************************************************/
int elimina_maquina (char *nombre_logico) {

   char valor_hash;       // Valor hash del parametro nombre logico
   t_maquina *aux_maq,    // variable auxiliar para buscar la mquina
			 *ant_maq;    // Maquina anterior a la que se quiere eliminar

   valor_hash = f_hash(nombre_logico);
   aux_maq = red[valor_hash];

   while ((aux_maq != NULL) && strcmp(aux_maq->logico, nombre_logico)) {
	  ant_maq = aux_maq;
	  aux_maq = aux_maq->siguiente;
   }
   if (aux_maq == NULL)
	  return (0);

   if (red[valor_hash] == aux_maq)
	  red[valor_hash] = aux_maq->siguiente;
   else
	  ant_maq->siguiente = aux_maq->siguiente;
   return(1);
}



/*************************************************************************
FUNCIONALIDAD: Cambia la direccin fsica de una mquina. Esta funcin
			   debe ser utilizada con mucho cuidado ya que la red podra
			   contener datos inconsistentes.
PARAMETROS:
    logico: nombre logico de la maquina
    fisico: direccin fisica de la maquina
DEVUELVE: Si se ha insertado, devuelve un 0. En caso contrario, se asume
		  la existencia de un error y devuelve -1.
/*************************************************************************/
int cambiar_dir_fis (char nombre_logico[], char dir_fis[]){

   char valor_hash;      // Valor hash del nombre lgico
   t_maquina *aux_maq;   // variable auxiliar para buscar la mquina

   valor_hash = f_hash(nombre_logico);
   aux_maq = red[valor_hash];

   while ((aux_maq != NULL) && strcmp(aux_maq->logico, nombre_logico))
	  aux_maq = aux_maq->siguiente;
   if (aux_maq == NULL)
	   return (0);
   else
	   memcpy (aux_maq->fisico, dir_fis, TAM_DIR);
   return (1);
}



/*************************************************************************
FUNCIONALIDAD: Construye el paquete que contiene:
			   datos enlace: Direcciones fisicas origen y destino y tipo
			   datagrama: nombre logico origen y destino, nmero del paquete
						  y paquetes totales, flags, tamao de los datos que
						  componen la informacin restante y dos variables
						  auxiliares para posibles necesidades futuras.
			   datos: informacin de niveles superiores.
PARAMETROS:
    datagrama: datos correspondientes a la estructura t_datagrama
				(ver fichero red.h)
    datos: informacin a ser enviada
    tam_paq: tamao final del paquete (direcciones fisicas origen y destino,
			  tipo, datagrama e informacin).
DEVUELVE: Una zona de memoria con el paquete construido. Si no se ha podido
		  construir, se devuelve NULL.
/*************************************************************************/
char *construye_paq (t_datagrama datagrama, char *datos, int *tam_paq) {

   char *paq;			          // Paquete final listo para ser enviado
   char *datag_datos,	          // Datagrama + datos
		*ptr;				      // Puntero auxiliar
   char dir_fis_dest [TAM_DIR];   // Direccin fisica destino

	// reservamos memoria para construir el paquete a ser enviado
	datag_datos = (char *) malloc (datagrama.tam_dat + sizeof(t_datagrama));
	if (datag_datos == NULL)
	   return (NULL);

	// copiamos y anexamos la informacin pasada por parametro
	ptr = datag_datos;
	memcpy (ptr, &(datagrama), sizeof (t_datagrama));
	ptr += sizeof (t_datagrama);
	memcpy(ptr, datos, datagrama.tam_dat);
	ptr += datagrama.tam_dat;
	*tam_paq = sizeof(t_datagrama) + datagrama.tam_dat;

	// obtenemos la direccin fisica buscando en las maquinas que la
	// se han encontrado en la red. Si no la conocemos, lo mandamos
	// en broadcast.
	if (buscar_maquina (datagrama.logico_dt, dir_fis_dest))
	   paq = make_pkt (dir_fis_dest, dir_origen, tipo, datag_datos, tam_paq);
	else
	   paq = make_pkt (BROAD_CAST, dir_origen, tipo, datag_datos, tam_paq);

	// Libero la zona donde almacenaba el datagrama y la informacin de
	// niveles superiores.
	free (datag_datos);

	return (paq);
}


/*************************************************************************
FUNCIONALIDAD: Funcion que envia un paquete el cual ha sido construido
			   y ya esta listo para ser enviado.
PARAMETROS:
    len : tamao final del paquete
    buffer: almacena los datos a enviar (dirs. fis, tipo, datagrama, datos)
DEVUELVE: Devuelve -1 si se ha producido algun error. En caso contrario,
		  devuelve 0.
/*************************************************************************/
int enviar_paq_constr (char *buffer, int len) {
struct REGPACK reg;

  return(send_pkt(buffer, len, &reg));
}


/*************************************************************************
FUNCIONALIDAD: Funcion que envia un paquete construyendo previamente a
			   partir del datagrama y de los datos (informacin de niveles
			   superiores). Si existe algun error (memoria insuficiente)
			   aborta el programa.
PARAMETROS:
    datagrama : ver fichero red.h
    datos: informacin de niveles superiores.
/*************************************************************************/
void enviar_paq (t_datagrama datagrama, char *datos) {

   char *paquete_a_enviar;
   int tam_paq;

   paquete_a_enviar = construye_paq (datagrama, datos, &tam_paq);
   if (paquete_a_enviar == NULL) {
	  fprintf (stderr, "\nMemoria insuficiente\n\n");
	  exit (-1);
   }
   if (enviar_paq_constr (paquete_a_enviar, tam_paq) == -1) {
	  fprintf(stderr, "\nError al enviar un paquete 'send_paket'\n\n");
	  exit (-1);
   }
   free(paquete_a_enviar);
}


/*************************************************************************
FUNCIONALIDAD: Obtiene el siguiente paquete mediante una poltica FIFO
			   que se ha almacen en el momento de su recepcin para
			   posteriormente tratarlo.
PARAMETROS:
    paquete: paquete que contiene la informacin del siguiente paquete
			  a tratar.
DEVUELVE: Si no existen ningn paquete por tratar devuelve un 0 y el valor
		  de paquete es indeterminado. En caso de que existan paquete,
		  devuelve un 1 y la informacin del paquete se recupera mediante
		  el parametro 'paquete'.
/*************************************************************************/
int sacar_paq (t_paquete *paquete) {

   if (!buffer_entr[sac_buffer_entr].libre) {
	 memcpy ((char *)paquete, (char *)buffer_entr[sac_buffer_entr].trama,
			 TAM_MAX_TRAMA);
	 buffer_entr[sac_buffer_entr].libre = 1;
	 sac_buffer_entr = (sac_buffer_entr + 1) % NUM_MAX_TRAMAS;
	 return (1);
   }
  return (0);
}



